#!/usr/bin/env python
# coding: utf-8

# # MLPs with Keras (IMDB & MNIST)

# ## Dependencies

# In[1]:


get_ipython().system('pip install pandas')
get_ipython().system('pip install numpy')
get_ipython().system('pip install matplotlib')
get_ipython().system('pip install tensorflow')
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf

get_ipython().run_line_magic('matplotlib', 'inline')


# In[10]:


get_ipython().system('pip install scikit-learn')


# We will use the IMDB dataset.
# 
# This is a dataset of 25,000 movies reviews from IMDB, labeled by sentiment (positive/negative). Reviews have been preprocessed, and each review is encoded as a list of word indexes (integers). For convenience, words are indexed by overall frequency in the dataset, so that for instance the integer "3" encodes the 3rd most frequent word in the data. This allows for quick filtering operations such as: "only consider the top 10,000 most common words, but eliminate the top 20 most common words".

# In[2]:


(x_train_imdb, y_train_imdb), (x_test_imdb, y_test_imdb) = tf.keras.datasets.imdb.load_data()

print(x_train_imdb[0]) 
print(y_train_imdb[0])

word_index = tf.keras.datasets.imdb.get_word_index()
index2word = dict((i + 3, word) for (word, i) in word_index.items())
index2word[0] = '[pad]'
index2word[1] = '[bos]' #Beginning of Sequence
index2word[2] = '[oov]'
x_train_imdb = np.array([' '.join([index2word[idx] for idx in text]) for text in x_train_imdb])
x_test_imdb = np.array([' '.join([index2word[idx] for idx in text]) for text in x_test_imdb]) 

print(x_train_imdb[0]) 
print(y_train_imdb[0])


# We will use our known binary representations

# In[11]:


from sklearn.feature_extraction.text import CountVectorizer

#min-df --> when building the vocabulary ignore terms that have a document frequency strictly lower than the given threshold.
binary_vectorizer = CountVectorizer(binary=True, min_df=100)
x_train_imdb_binary = binary_vectorizer.fit_transform(x_train_imdb)
x_test_imdb_binary = binary_vectorizer.transform(x_test_imdb)
print('Vocabulary size:', len(binary_vectorizer.vocabulary_))


# In[12]:


type(x_train_imdb_binary[0])


# In[13]:


x_train_imdb_binary = x_train_imdb_binary.toarray()
x_test_imdb_binary = x_test_imdb_binary.toarray()


# In[14]:


type(x_train_imdb_binary[0])


# ## MLPs

# We can add hidden layers in order to create an MLP and train it with the binary vectors

# In[16]:


from tensorflow.keras.utils import plot_model
from IPython.display import Image 

imdb_mlp = tf.keras.models.Sequential([
      tf.keras.layers.Input(shape=(3834,)),
      tf.keras.layers.Dense(units=256, activation='relu'),
      tf.keras.layers.Dropout(rate=0.2),
      tf.keras.layers.Dense(units=128, activation='relu'),
      tf.keras.layers.Dense(units=64, activation='relu'),
      tf.keras.layers.Dropout(rate=0.5),
      tf.keras.layers.Dense(units=1, activation='sigmoid')
])

plot_model(imdb_mlp, to_file='model.png', show_shapes=True)
Image('model.png')


# In[23]:


imdb_mlp.compile(loss=tf.keras.losses.BinaryCrossentropy(), 
                  optimizer=tf.keras.optimizers.SGD(),
                  metrics=['binary_accuracy'])
imdb_mlp.fit(x=x_train_imdb_binary, y=y_train_imdb,
             epochs=10, verbose=1, batch_size=32)


# In[24]:


print(imdb_mlp.evaluate(x_test_imdb_binary, y_test_imdb))


# ### Experiment with the MNIST dataset using the Keras' Functional API 
# 
# This is a dataset of 60,000 28x28 grayscale images of the 10 digits, along with a test set of 10,000 images

# In[17]:


(x_train_mnist, y_train_mnist), (x_test_mnist, y_test_mnist) = tf.keras.datasets.mnist.load_data()
print(x_train_mnist[0].shape)
print(x_train_mnist[0])
x_train_mnist = x_train_mnist.astype("float32") / 255
x_test_mnist = x_test_mnist.astype("float32") / 255 
print(x_train_mnist[0])


# In[32]:


print(x_train_mnist[0].shape)
print(y_train_mnist[0])


# In[33]:


fig = plt.figure(figsize=(8, 8))
for i in range(25):
  plt.subplot(5, 5, i+1)
  plt.tight_layout()
  plt.imshow(x_train_mnist[i], cmap='gray', interpolation='none')
  plt.title("Digit: {}".format(y_train_mnist[i]))
  plt.xticks(list())
  plt.yticks(list())


# In[34]:


np.unique(y_train_mnist)


# In[35]:


def get_mnist_mlp(num_classes):
  inp = tf.keras.layers.Input(shape=(28, 28), name='inputs')
  x = tf.keras.layers.Flatten(name='flatten_inputs')(inp) # 28 * 28 = 784
  x = tf.keras.layers.Dense(units=512, activation='relu', name='hidden_1')(x)
  x = tf.keras.layers.Dense(units=256, activation='relu', name='hidden_2')(x)
  x = tf.keras.layers.Dense(units=num_classes, activation='softmax', 
                            name='classifier')(x)
  return tf.keras.models.Model(inputs=inp, outputs=x, name='mnist_mlp')

mnist_mlp = get_mnist_mlp(len(np.unique(y_train_mnist)))
plot_model(mnist_mlp, to_file='model.png', show_shapes=True)
Image('model.png')


# In[40]:


mnist_mlp.compile(loss=tf.keras.losses.CategoricalCrossentropy(), 
                  metrics=['accuracy'], optimizer=tf.keras.optimizers.SGD())
y_train_mnist = tf.keras.utils.to_categorical(y_train_mnist, num_classes=10, 
                                              dtype="int32")
y_test_mnist = tf.keras.utils.to_categorical(y_test_mnist, num_classes=10, 
                                             dtype="int32")
y_train_mnist[0]


# In[42]:


mnist_mlp_history = mnist_mlp.fit(x_train_mnist, y_train_mnist, batch_size=32, 
                                  epochs=10, validation_split=0.2)


# In[51]:


def plot(his, kind):
  train = his.history[kind]
  val = his.history['val_' + kind]
  epochs = range(1, len(train)+1)
  plt.figure(figsize=(12,9))
  plt.plot(epochs, train, 'b', label='Training ' + kind)
  plt.plot(epochs, val, 'orange', label='Validation ' + kind)
  plt.title('Training and validation ' + kind) 
  plt.xlabel('Epochs')
  plt.ylabel(kind)
  # x1, x2, y1, y2 = plt.axis()  
  # plt.axis((x1, x2, 0, 0.2))
  plt.legend()
  plt.show()


# In[52]:


plot(mnist_mlp_history, 'loss')


# In[61]:


print(mnist_mlp.evaluate(x_test_mnist, y_test_mnist))


# In[60]:


y_pred_mnist = mnist_mlp.predict(x_test_mnist)
print(y_pred_mnist[0])
y_pred_mnist = np.argmax(y_pred_mnist, axis=-1)
print(y_pred_mnist[0])
y_pred_mnist = tf.keras.utils.to_categorical(y_pred_mnist, 
                                             num_classes=10, dtype="int32")
print(y_pred_mnist[0])
# Print evaluation report
print(metrics.classification_report(y_test_mnist, y_pred_mnist))


# In[59]:


fig = plt.figure(figsize=(8, 8))
for i in range(25):
  plt.subplot(5, 5, i+1)
  plt.tight_layout()
  plt.imshow(x_test_mnist[i], cmap='gray', interpolation='none')
  plt.title(f"G: {np.argmax(y_test_mnist[i])} P: {np.argmax(y_pred_mnist[i])}")
  plt.xticks(list())
  plt.yticks(list())

